searchData={"items":[{"type":"behaviour","title":"Hammer","doc":"Hammer is a rate-limiting library for Elixir.\n\nIt provides a simple way for creating rate limiters, and comes with a built-in ETS backend.\n\n    defmodule MyApp.RateLimit do\n      use Hammer, backend: :ets\n    end\n\n    # Start the rate limiter, in case of ETS it will create the ETS table and schedule cleanups\n    MyApp.RateLimit.start_link(clean_period: :timer.minutes(10))\n\n    # Check the rate limit allowing 10 requests per second\n    MyApp.RateLimit.hit(\"some-key\", _scale = :timer.seconds(1), _limit = 10)","ref":"Hammer.html"},{"type":"macro","title":"Hammer.__using__/1","doc":"Use the Hammer library in a module to create a rate limiter.\n\n    defmodule MyApp.RateLimit do\n      use Hammer, backend: :ets\n    end","ref":"Hammer.html#__using__/1"},{"type":"callback","title":"Hammer.get/2","doc":"Optional callback for getting the counter value for a key.\n\nReturns the current counter value.","ref":"Hammer.html#c:get/2"},{"type":"callback","title":"Hammer.hit/3","doc":"Checks if a key is allowed to perform an action, and increment the counter.\n\nSame as `hit/4` with `increment` set to 1.","ref":"Hammer.html#c:hit/3"},{"type":"callback","title":"Hammer.hit/4","doc":"Optional callback to check if a key is allowed to perform an action, and increment the counter.\n\nReturns `{:allow, count}` if the action is allowed, or `{:deny, timeout}` if the action is denied.\n\nThis is the only required callback.","ref":"Hammer.html#c:hit/4"},{"type":"callback","title":"Hammer.inc/2","doc":"Same as `inc/3` with `increment` set to 1.","ref":"Hammer.html#c:inc/2"},{"type":"callback","title":"Hammer.inc/3","doc":"Optional callback for incrementing a counter value for a kit without performing limit check.\n\nReturns the new counter value.","ref":"Hammer.html#c:inc/3"},{"type":"callback","title":"Hammer.set/3","doc":"Optional callback for setting the counter value for a key.\n\nReturns the new counter value.","ref":"Hammer.html#c:set/3"},{"type":"type","title":"Hammer.count/0","doc":"","ref":"Hammer.html#t:count/0"},{"type":"type","title":"Hammer.increment/0","doc":"","ref":"Hammer.html#t:increment/0"},{"type":"type","title":"Hammer.key/0","doc":"","ref":"Hammer.html#t:key/0"},{"type":"type","title":"Hammer.limit/0","doc":"","ref":"Hammer.html#t:limit/0"},{"type":"type","title":"Hammer.scale/0","doc":"","ref":"Hammer.html#t:scale/0"},{"type":"module","title":"Hammer.Atomic","doc":"A rate limiter implementation using Erlang's :atomics module for atomic counters.\n\nThis provides fast, atomic counter operations without the overhead of ETS or process messaging.\nRequires Erlang/OTP 21.2 or later.","ref":"Hammer.Atomic.html"},{"type":"function","title":"Hammer.Atomic.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Hammer.Atomic.html#child_spec/1"},{"type":"function","title":"Hammer.Atomic.now/0","doc":"Returns the current time in milliseconds.","ref":"Hammer.Atomic.html#now/0"},{"type":"function","title":"Hammer.Atomic.start_link/1","doc":"Starts the atomic rate limiter process.\n\nOptions:\n- `:clean_period` - How often to run cleanup (ms). Default 1 minute.\n- `:key_older_than` - Max age for entries (ms). Default 24 hours.","ref":"Hammer.Atomic.html#start_link/1"},{"type":"type","title":"Hammer.Atomic.config/0","doc":"","ref":"Hammer.Atomic.html#t:config/0"},{"type":"type","title":"Hammer.Atomic.start_option/0","doc":"","ref":"Hammer.Atomic.html#t:start_option/0"},{"type":"module","title":"Hammer.Atomic.FixWindow","doc":"This module implements the Fix Window algorithm.\n\nThe fixed window algorithm works by dividing time into fixed intervals or \"windows\"\nof a specified duration (scale). Each window tracks request counts independently.\n\nFor example, with a 60 second window:\n- Window 1: 0-60 seconds\n- Window 2: 60-120 seconds\n- And so on...\n\nThe algorithm:\n1. When a request comes in, we:\n   - Calculate which window it belongs to based on current time\n   - Increment the counter for that window\n   - Store expiration time as end of window\n2. To check if rate limit is exceeded:\n   - If count <= limit: allow request\n   - If count > limit: deny and return time until window expires\n3. Old windows are automatically cleaned up after expiration\n\nThis provides simple rate limiting but has edge cases where a burst of requests\nspanning a window boundary could allow up to 2x the limit in a short period.\nFor more precise limiting, consider using the sliding window algorithm instead.\n\nThe fixed window algorithm is a good choice when:\n\n- You need simple, predictable rate limiting with clear time boundaries\n- The exact precision of the rate limit is not critical\n- You want efficient implementation with minimal storage overhead\n- Your use case can tolerate potential bursts at window boundaries\n\nCommon use cases include:\n\n- Basic API rate limiting where occasional bursts are acceptable\n- Protecting backend services from excessive load\n- Implementing fair usage policies\n- Scenarios where clear time-based quotas are desired (e.g. \"100 requests per minute\")\n\nThe main tradeoff is that requests near window boundaries can allow up to 2x the\nintended limit in a short period. For example with a limit of 100 per minute:\n- 100 requests at 11:59:59\n- Another 100 requests at 12:00:01\n\nThis results in 200 requests in 2 seconds, while still being within limits.\nIf this behavior is problematic, consider using the sliding window algorithm instead.\n\nThe fixed window algorithm supports the following options:\n\n- `:clean_period` - How often to run the cleanup process (in milliseconds)\n  Defaults to 1 minute. The cleanup process removes expired window entries.\n\n\nExample configuration:\n\n    MyApp.RateLimit.start_link(\n      clean_period: :timer.minutes(5),\n    )\n\nThis would run cleanup every 5 minutes and clean up old windows.","ref":"Hammer.Atomic.FixWindow.html"},{"type":"function","title":"Hammer.Atomic.FixWindow.get/3","doc":"Returns the count of requests for a given key within the last   seconds.","ref":"Hammer.Atomic.FixWindow.html#get/3"},{"type":"function","title":"Hammer.Atomic.FixWindow.hit/5","doc":"Checks if a key is allowed to perform an action based on the fixed window algorithm.","ref":"Hammer.Atomic.FixWindow.html#hit/5"},{"type":"function","title":"Hammer.Atomic.FixWindow.inc/4","doc":"Increments the counter for a given key in the fixed window algorithm.","ref":"Hammer.Atomic.FixWindow.html#inc/4"},{"type":"function","title":"Hammer.Atomic.FixWindow.set/4","doc":"Sets the counter for a given key in the fixed window algorithm.","ref":"Hammer.Atomic.FixWindow.html#set/4"},{"type":"module","title":"Hammer.Atomic.LeakyBucket","doc":"This module implements the Leaky Bucket algorithm.\n\nThe leaky bucket algorithm works by modeling a bucket that:\n- Fills up with requests at the input rate\n- \"Leaks\" requests at a constant rate\n- Has a maximum capacity (the bucket size)\n\nFor example, with a leak rate of 10 requests/second and bucket size of 100:\n- Requests add to the bucket's current level\n- The bucket leaks 10 requests per second steadily\n- If bucket reaches capacity (100), new requests are denied\n- Once bucket level drops, new requests are allowed again\n\nThe algorithm:\n1. When a request comes in, we:\n   - Calculate how much has leaked since last request\n   - Subtract leaked amount from current bucket level\n   - Try to add new request to bucket\n   - Store new bucket level and timestamp\n2. To check if rate limit is exceeded:\n   - If new bucket level <= capacity: allow request\n   - If new bucket level > capacity: deny and return time until enough leaks\n3. Old entries are automatically cleaned up after expiration\n\nThis provides smooth rate limiting with ability to handle bursts up to bucket size.\nThe leaky bucket is a good choice when:\n\n- You need to enforce a constant processing rate\n- Want to allow temporary bursts within bucket capacity\n- Need to smooth out traffic spikes\n- Want to prevent resource exhaustion\n\nCommon use cases include:\n\n- API rate limiting needing consistent throughput\n- Network traffic shaping\n- Service protection from sudden load spikes\n- Queue processing rate control\n- Scenarios needing both burst tolerance and steady-state limits\n\nThe main advantages are:\n- Smooth, predictable output rate\n- Configurable burst tolerance\n- Natural queueing behavior\n\nThe tradeoffs are:\n- More complex implementation than fixed windows\n- Need to track last request time and current bucket level\n- May need tuning of bucket size and leak rate parameters\n\nFor example, with 100 requests/sec limit and 500 bucket size:\n- Can handle bursts of up to 500 requests\n- But long-term average rate won't exceed 100/sec\n- Provides smoother traffic than fixed windows\n\nThe leaky bucket algorithm supports the following options:\n\n- `:clean_period` - How often to run the cleanup process (in milliseconds)\n  Defaults to 1 minute. The cleanup process removes expired bucket entries.\n\n- `:key_older_than` - Optional maximum age for bucket entries (in milliseconds)\n  If set, entries older than this will be removed during cleanup.\n  This helps prevent memory growth from abandoned buckets.\n\nExample configuration:\n\n    MyApp.RateLimit.start_link(\n      clean_period: :timer.minutes(5),\n      key_older_than: :timer.hours(24)\n    )\n\nThis would run cleanup every 5 minutes and remove buckets not used in 24 hours.","ref":"Hammer.Atomic.LeakyBucket.html"},{"type":"function","title":"Hammer.Atomic.LeakyBucket.get/2","doc":"Returns the current level of the bucket for a given key.","ref":"Hammer.Atomic.LeakyBucket.html#get/2"},{"type":"function","title":"Hammer.Atomic.LeakyBucket.hit/5","doc":"Checks if a key is allowed to perform an action, and increment the counter by the given amount.","ref":"Hammer.Atomic.LeakyBucket.html#hit/5"},{"type":"module","title":"Hammer.Atomic.TokenBucket","doc":"This module implements the Token Bucket algorithm.\nThe token bucket algorithm works by modeling a bucket that:\n- Fills with tokens at a constant rate (the refill rate)\n- Has a maximum capacity of tokens (the bucket size)\n- Each request consumes one or more tokens\n- If there are enough tokens, the request is allowed\n- If not enough tokens, the request is denied\n\nFor example, with a refill rate of 10 tokens/second and bucket size of 100:\n- Tokens are added at 10 per second up to max of 100\n- Each request needs tokens to proceed\n- If bucket has enough tokens, request allowed and tokens consumed\n- If not enough tokens, request denied until bucket refills\n\nThe algorithm:\n1. When a request comes in, we:\n   - Calculate tokens added since last request based on time elapsed\n   - Add new tokens to bucket (up to max capacity)\n   - Try to consume tokens for the request\n   - Store new token count and timestamp\n2. To check if rate limit is exceeded:\n   - If enough tokens: allow request and consume tokens\n   - If not enough: deny and return time until enough tokens refill\n3. Old entries are automatically cleaned up after expiration\n\nThis provides smooth rate limiting with ability to handle bursts up to bucket size.\nThe token bucket is a good choice when:\n\n- You need to allow temporary bursts of traffic\n- Want to enforce an average rate limit\n- Need to support different costs for different operations\n- Want to avoid the sharp edges of fixed windows\n\nCommon use cases include:\n\n- API rate limiting with burst tolerance\n- Network traffic shaping\n- Resource allocation control\n- Gaming systems with \"energy\" mechanics\n- Scenarios needing flexible rate limits\n\nThe main advantages are:\n- Natural handling of bursts\n- Flexible token costs for different operations\n- Smooth rate limiting behavior\n- Simple to reason about\n\nThe tradeoffs are:\n- Need to track token count and last update time\n- May need tuning of bucket size and refill rate\n- More complex than fixed windows\n\nFor example with 100 tokens/minute limit and 500 bucket size:\n- Can handle bursts using saved up tokens\n- Automatically smooths out over time\n- Different operations can cost different amounts\n- More flexible than fixed request counts\n\nThe token bucket algorithm supports the following options:\n\n- `:clean_period` - How often to run the cleanup process (in milliseconds)\n  Defaults to 1 minute. The cleanup process removes expired bucket entries.\n\n- `:key_older_than` - Optional maximum age for bucket entries (in milliseconds)\n  If set, entries older than this will be removed during cleanup.\n  This helps prevent memory growth from abandoned buckets.\n\nExample configuration:\n\n    MyApp.RateLimit.start_link(\n      clean_period: :timer.minutes(5),\n      key_older_than: :timer.hours(24)\n    )\n\nThis would run cleanup every 5 minutes and remove buckets not used in 24 hours.","ref":"Hammer.Atomic.TokenBucket.html"},{"type":"function","title":"Hammer.Atomic.TokenBucket.get/2","doc":"Returns the current level of the bucket for a given key.","ref":"Hammer.Atomic.TokenBucket.html#get/2"},{"type":"module","title":"Hammer.ETS","doc":"An ETS backend for Hammer.\n\nTo use the ETS backend, you need to start the process that creates and cleans the ETS table. The table is named after the module.\n\n    defmodule MyApp.RateLimit do\n      use Hammer, backend: :ets\n    end\n\n    MyApp.RateLimit.start_link(clean_period: :timer.minutes(1))\n\nRuntime configuration:\n- `:clean_period` - (in milliseconds) period to clean up expired entries, defaults to 1 minute\n- `:key_older_than` - (in milliseconds) maximum age for entries before they are cleaned up, defaults to 1 hour\n- `:algorithm` - the rate limiting algorithm to use, one of: `:fix_window`, `:sliding_window`, `:leaky_bucket`, `:token_bucket`. Defaults to `:fix_window`","ref":"Hammer.ETS.html"},{"type":"function","title":"Hammer.ETS.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Hammer.ETS.html#child_spec/1"},{"type":"function","title":"Hammer.ETS.now/0","doc":"","ref":"Hammer.ETS.html#now/0"},{"type":"function","title":"Hammer.ETS.start_link/1","doc":"Starts the process that creates and cleans the ETS table.\n\nAccepts the following options:\n  - `:clean_period` - How often to run the cleanup process (in milliseconds). Defaults to 1 minute.\n  - `:key_older_than` - Optional maximum age for bucket entries (in milliseconds). Defaults to 24 hours.\n    Entries older than this will be removed during cleanup.\n  - `:algorithm` - The rate limiting algorithm to use. Can be `:fixed_window`, `:sliding_window`,\n    `:token_bucket`, or `:leaky_bucket`. Defaults to `:fixed_window`.\n  - optional `:debug`, `:spawn_opts`, and `:hibernate_after` GenServer options","ref":"Hammer.ETS.html#start_link/1"},{"type":"function","title":"Hammer.ETS.update_counter/4","doc":"","ref":"Hammer.ETS.html#update_counter/4"},{"type":"type","title":"Hammer.ETS.config/0","doc":"","ref":"Hammer.ETS.html#t:config/0"},{"type":"type","title":"Hammer.ETS.start_option/0","doc":"","ref":"Hammer.ETS.html#t:start_option/0"},{"type":"module","title":"Hammer.ETS.FixWindow","doc":"This module implements the Fix Window algorithm.\n\nThe fixed window algorithm works by dividing time into fixed intervals or \"windows\"\nof a specified duration (scale). Each window tracks request counts independently.\n\nFor example, with a 60 second window:\n- Window 1: 0-60 seconds\n- Window 2: 60-120 seconds\n- And so on...\n\nThe algorithm:\n1. When a request comes in, we:\n   - Calculate which window it belongs to based on current time\n   - Increment the counter for that window\n   - Store expiration time as end of window\n2. To check if rate limit is exceeded:\n   - If count <= limit: allow request\n   - If count > limit: deny and return time until window expires\n3. Old windows are automatically cleaned up after expiration\n\nThis provides simple rate limiting but has edge cases where a burst of requests\nspanning a window boundary could allow up to 2x the limit in a short period.\nFor more precise limiting, consider using the sliding window algorithm instead.\n\nThe fixed window algorithm is a good choice when:\n\n- You need simple, predictable rate limiting with clear time boundaries\n- The exact precision of the rate limit is not critical\n- You want efficient implementation with minimal storage overhead\n- Your use case can tolerate potential bursts at window boundaries\n\nCommon use cases include:\n\n- Basic API rate limiting where occasional bursts are acceptable\n- Protecting backend services from excessive load\n- Implementing fair usage policies\n- Scenarios where clear time-based quotas are desired (e.g. \"100 requests per minute\")\n\nThe main tradeoff is that requests near window boundaries can allow up to 2x the\nintended limit in a short period. For example with a limit of 100 per minute:\n- 100 requests at 11:59:59\n- Another 100 requests at 12:00:01\n\nThis results in 200 requests in 2 seconds, while still being within limits.\nIf this behavior is problematic, consider using the sliding window algorithm instead.\n\nThe fixed window algorithm supports the following options:\n\n- `:clean_period` - How often to run the cleanup process (in milliseconds)\n  Defaults to 1 minute. The cleanup process removes expired window entries.\n\n\nExample configuration:\n\n    MyApp.RateLimit.start_link(\n      clean_period: :timer.minutes(5),\n    )\n\nThis would run cleanup every 5 minutes and clean up old windows.","ref":"Hammer.ETS.FixWindow.html"},{"type":"function","title":"Hammer.ETS.FixWindow.clean/1","doc":"Cleans up all of the old entries from the table.","ref":"Hammer.ETS.FixWindow.html#clean/1"},{"type":"function","title":"Hammer.ETS.FixWindow.get/3","doc":"Returns the count of requests for a given key within the last   seconds.","ref":"Hammer.ETS.FixWindow.html#get/3"},{"type":"function","title":"Hammer.ETS.FixWindow.hit/5","doc":"Checks if a key is allowed to perform an action based on the fixed window algorithm.","ref":"Hammer.ETS.FixWindow.html#hit/5"},{"type":"function","title":"Hammer.ETS.FixWindow.inc/4","doc":"Increments the counter for a given key in the fixed window algorithm.","ref":"Hammer.ETS.FixWindow.html#inc/4"},{"type":"function","title":"Hammer.ETS.FixWindow.set/4","doc":"Sets the counter for a given key in the fixed window algorithm.","ref":"Hammer.ETS.FixWindow.html#set/4"},{"type":"module","title":"Hammer.ETS.LeakyBucket","doc":"This module implements the Leaky Bucket algorithm.\n\nThe leaky bucket algorithm works by modeling a bucket that:\n- Fills up with requests at the input rate\n- \"Leaks\" requests at a constant rate\n- Has a maximum capacity (the bucket size)\n\nFor example, with a leak rate of 10 requests/second and bucket size of 100:\n- Requests add to the bucket's current level\n- The bucket leaks 10 requests per second steadily\n- If bucket reaches capacity (100), new requests are denied\n- Once bucket level drops, new requests are allowed again\n\nThe algorithm:\n1. When a request comes in, we:\n   - Calculate how much has leaked since last request\n   - Subtract leaked amount from current bucket level\n   - Try to add new request to bucket\n   - Store new bucket level and timestamp\n2. To check if rate limit is exceeded:\n   - If new bucket level <= capacity: allow request\n   - If new bucket level > capacity: deny and return time until enough leaks\n3. Old entries are automatically cleaned up after expiration\n\nThis provides smooth rate limiting with ability to handle bursts up to bucket size.\nThe leaky bucket is a good choice when:\n\n- You need to enforce a constant processing rate\n- Want to allow temporary bursts within bucket capacity\n- Need to smooth out traffic spikes\n- Want to prevent resource exhaustion\n\nCommon use cases include:\n\n- API rate limiting needing consistent throughput\n- Network traffic shaping\n- Service protection from sudden load spikes\n- Queue processing rate control\n- Scenarios needing both burst tolerance and steady-state limits\n\nThe main advantages are:\n- Smooth, predictable output rate\n- Configurable burst tolerance\n- Natural queueing behavior\n\nThe tradeoffs are:\n- More complex implementation than fixed windows\n- Need to track last request time and current bucket level\n- May need tuning of bucket size and leak rate parameters\n\nFor example, with 100 requests/sec limit and 500 bucket size:\n- Can handle bursts of up to 500 requests\n- But long-term average rate won't exceed 100/sec\n- Provides smoother traffic than fixed windows\n\nThe leaky bucket algorithm supports the following options:\n\n- `:clean_period` - How often to run the cleanup process (in milliseconds)\n  Defaults to 1 minute. The cleanup process removes expired bucket entries.\n\n- `:key_older_than` - Optional maximum age for bucket entries (in milliseconds)\n  If set, entries older than this will be removed during cleanup.\n  This helps prevent memory growth from abandoned buckets.\n\nExample configuration:\n\n    MyApp.RateLimit.start_link(\n      clean_period: :timer.minutes(5),\n      key_older_than: :timer.hours(24)\n    )\n\nThis would run cleanup every 5 minutes and remove buckets not used in 24 hours.","ref":"Hammer.ETS.LeakyBucket.html"},{"type":"function","title":"Hammer.ETS.LeakyBucket.clean/1","doc":"Cleans up all of the old entries from the table based on the `key_older_than` option.","ref":"Hammer.ETS.LeakyBucket.html#clean/1"},{"type":"function","title":"Hammer.ETS.LeakyBucket.get/2","doc":"Returns the current level of the bucket for a given key.","ref":"Hammer.ETS.LeakyBucket.html#get/2"},{"type":"function","title":"Hammer.ETS.LeakyBucket.hit/5","doc":"Checks if a key is allowed to perform an action, and increment the counter by the given amount.","ref":"Hammer.ETS.LeakyBucket.html#hit/5"},{"type":"module","title":"Hammer.ETS.SlidingWindow","doc":"This module implements the Rate Limiting Sliding Window algorithm.\n\nThe sliding window algorithm works by tracking requests within a moving time window.\nUnlike a fixed window that resets at specific intervals, the sliding window\nprovides a smoother rate limiting experience by considering the most recent\nwindow of time.\n\nFor example, with a 60 second window:\n- At time t, we look back 60 seconds and count all requests in that period\n- At time t+1, we look back 60 seconds from t+1, dropping any requests older than that\n- This creates a \"sliding\" effect where the window gradually moves forward in time\n\nThe algorithm:\n1. When a request comes in, we store it with the current timestamp\n2. To check if rate limit is exceeded, we:\n   - Count all requests within the last   seconds\n   - If count <= limit: allow the request\n   - If count > limit: deny and return time until oldest request expires\n3. Old entries outside the window are automatically cleaned up\n\nThis provides more precise rate limiting compared to fixed windows, avoiding\nthe edge case where a burst of requests spans a fixed window boundary.\n\nThe sliding window algorithm is a good choice when:\n\n- You need precise rate limiting without allowing bursts at window boundaries\n- Accuracy of the rate limit is critical for your application\n- You can accept slightly higher storage overhead compared to fixed windows\n- You want to avoid sudden changes in allowed request rates\n\nCommon use cases include:\n\n- API rate limiting where consistent request rates are important\n- Financial transaction rate limiting\n- User action throttling requiring precise control\n- Gaming or real-time applications needing smooth rate control\n- Security-sensitive rate limiting scenarios\n\nThe main advantages over fixed windows are:\n\n- No possibility of 2x burst at window boundaries\n- Smoother rate limiting behavior\n- More predictable request patterns\n\nThe tradeoffs are:\n- Slightly more complex implementation\n- Higher storage requirements (need to store individual request timestamps)\n- More computation required to check limits (need to count requests in window)\n\nFor example, with a limit of 100 requests per minute:\n- Fixed window might allow 200 requests across a boundary (100 at 11:59, 100 at 12:00)\n- Sliding window ensures no more than 100 requests in ANY 60 second period\n\nThe sliding window algorithm supports the following options:\n\n- `:clean_period` - How often to run the cleanup process (in milliseconds)\n  Defaults to 1 minute. The cleanup process removes expired window entries.\n\nExample configuration:\n\n    MyApp.RateLimit.start_link(\n      clean_period: :timer.minutes(5),\n    )\n\nThis would run cleanup every 5 minutes and clean up old windows.","ref":"Hammer.ETS.SlidingWindow.html"},{"type":"function","title":"Hammer.ETS.SlidingWindow.clean/1","doc":"Cleans up all of the old entries from the table based on the `key_older_than` option.","ref":"Hammer.ETS.SlidingWindow.html#clean/1"},{"type":"function","title":"Hammer.ETS.SlidingWindow.get/3","doc":"Returns the count of requests for a given key","ref":"Hammer.ETS.SlidingWindow.html#get/3"},{"type":"function","title":"Hammer.ETS.SlidingWindow.hit/4","doc":"Checks if a key is allowed to perform an action based on the sliding window algorithm.","ref":"Hammer.ETS.SlidingWindow.html#hit/4"},{"type":"module","title":"Hammer.ETS.TokenBucket","doc":"This module implements the Token Bucket algorithm.\nThe token bucket algorithm works by modeling a bucket that:\n- Fills with tokens at a constant rate (the refill rate)\n- Has a maximum capacity of tokens (the bucket size)\n- Each request consumes one or more tokens\n- If there are enough tokens, the request is allowed\n- If not enough tokens, the request is denied\n\nFor example, with a refill rate of 10 tokens/second and bucket size of 100:\n- Tokens are added at 10 per second up to max of 100\n- Each request needs tokens to proceed\n- If bucket has enough tokens, request allowed and tokens consumed\n- If not enough tokens, request denied until bucket refills\n\nThe algorithm:\n1. When a request comes in, we:\n   - Calculate tokens added since last request based on time elapsed\n   - Add new tokens to bucket (up to max capacity)\n   - Try to consume tokens for the request\n   - Store new token count and timestamp\n2. To check if rate limit is exceeded:\n   - If enough tokens: allow request and consume tokens\n   - If not enough: deny and return time until enough tokens refill\n3. Old entries are automatically cleaned up after expiration\n\nThis provides smooth rate limiting with ability to handle bursts up to bucket size.\nThe token bucket is a good choice when:\n\n- You need to allow temporary bursts of traffic\n- Want to enforce an average rate limit\n- Need to support different costs for different operations\n- Want to avoid the sharp edges of fixed windows\n\nCommon use cases include:\n\n- API rate limiting with burst tolerance\n- Network traffic shaping\n- Resource allocation control\n- Gaming systems with \"energy\" mechanics\n- Scenarios needing flexible rate limits\n\nThe main advantages are:\n- Natural handling of bursts\n- Flexible token costs for different operations\n- Smooth rate limiting behavior\n- Simple to reason about\n\nThe tradeoffs are:\n- Need to track token count and last update time\n- May need tuning of bucket size and refill rate\n- More complex than fixed windows\n\nFor example with 100 tokens/minute limit and 500 bucket size:\n- Can handle bursts using saved up tokens\n- Automatically smooths out over time\n- Different operations can cost different amounts\n- More flexible than fixed request counts\n\nThe token bucket algorithm supports the following options:\n\n- `:clean_period` - How often to run the cleanup process (in milliseconds)\n  Defaults to 1 minute. The cleanup process removes expired bucket entries.\n\n- `:key_older_than` - Optional maximum age for bucket entries (in milliseconds)\n  If set, entries older than this will be removed during cleanup.\n  This helps prevent memory growth from abandoned buckets.\n\nExample configuration:\n\n    MyApp.RateLimit.start_link(\n      clean_period: :timer.minutes(5),\n      key_older_than: :timer.hours(24)\n    )\n\nThis would run cleanup every 5 minutes and remove buckets not used in 24 hours.","ref":"Hammer.ETS.TokenBucket.html"},{"type":"function","title":"Hammer.ETS.TokenBucket.clean/1","doc":"Cleans up all of the old entries from the table based on the `key_older_than` option.","ref":"Hammer.ETS.TokenBucket.html#clean/1"},{"type":"function","title":"Hammer.ETS.TokenBucket.get/2","doc":"Returns the current level of the bucket for a given key.","ref":"Hammer.ETS.TokenBucket.html#get/2"},{"type":"extras","title":"Changelog","doc":"# Changelog","ref":"changelog.html"},{"type":"extras","title":"7.0.0-rc.2 - 2024-12-17 - Changelog","doc":"- Fix type specs for ETS backends\n- Adds Atomic backends and possible algorithms\n- Added `:algorithm` option to the Atomic backend with support for:\n  - `:fix_window` (default) - Fixed time window rate limiting\n  - `:leaky_bucket` - Constant rate limiting with burst capacity\n  - `:token_bucket` - Token-based rate limiting with burst capacity\n- Add benchmarks file and run them with `bench`","ref":"changelog.html#7-0-0-rc-2-2024-12-17"},{"type":"extras","title":"7.0.0-rc.1 - 2024-12-13 - Changelog","doc":"- Improved API a little more. Should be compatibe with previous RC\n  - Made ETS backend more flexible with `:algorithm` option\n  - Added `:key_older_than` option to the ETS backend\n- Added `:algorithm` option to the ETS backend with support for:\n  - `:fix_window` (default) - Fixed time window rate limiting\n  - `:sliding_window` - Sliding time window for smoother rate limiting\n  - `:leaky_bucket` - Constant rate limiting with burst capacity\n  - `:token_bucket` - Token-based rate limiting with burst capacity","ref":"changelog.html#7-0-0-rc-1-2024-12-13"},{"type":"extras","title":"7.0.0-rc.0 - 2024-12-13 - Changelog","doc":"- Breaking change. Completely new API. Consider upgrading if you are experiencing performance or usability problems with Hammer v6. See [./guides/upgrade-v7.md] for upgrade instructions. https://github.com/ExHammer/hammer/pull/104\n- Hammer.Plug has been removed. See documentation for using Hammer as a plug in Phoenix.","ref":"changelog.html#7-0-0-rc-0-2024-12-13"},{"type":"extras","title":"6.2.1 - 2024-02-23 - Changelog","doc":"- Fix issue in OTP 26 and Elixir 1.15 by not using to_existing_atom in configuration","ref":"changelog.html#6-2-1-2024-02-23"},{"type":"extras","title":"Changed - Changelog","doc":"","ref":"changelog.html#changed"},{"type":"extras","title":"6.2.0 - 2024-01-31 - Changelog","doc":"- Ensure Elixir version is ~> 1.13 https://github.com/ExHammer/hammer/pull/79.","ref":"changelog.html#6-2-0-2024-01-31"},{"type":"extras","title":"6.1.0 - 2022-06-13 - Changelog","doc":"","ref":"changelog.html#6-1-0-2022-06-13"},{"type":"extras","title":"Changed - Changelog","doc":"- Updgrade dependency packages\n- Merged https://github.com/ExHammer/hammer/pull/41 resulting in ETC without GenServer (and therefore better performance)\n- Merged https://github.com/ExHammer/hammer/pull/46 remove additional whitespace\n- Updated Docs based on https://github.com/ExHammer/hammer/pull/45\n- Adds CREDITS.md","ref":"changelog.html#changed"},{"type":"extras","title":"6.0.0 - 2018-10-12 - Changelog","doc":"","ref":"changelog.html#6-0-0-2018-10-12"},{"type":"extras","title":"Changed - Changelog","doc":"- Change the `ETS` backend to throw an error if either `expiry_ms` or\n  `cleanup_interval_ms` config values are missing. This should have been fixed\n  ages ago.\n- Default `:pool_max_overflow` changed to `0`. It's a better default, given\n  that some users have seen weird errors when using a higher overflow.\n  In general, capacity should be increased by using a higher `:pool_size` instead\n- Changed how the ETS backend does cleanups of data, should be more performant.","ref":"changelog.html#changed"},{"type":"extras","title":"5.0.0 - 2018-05-18 - Changelog","doc":"","ref":"changelog.html#5-0-0-2018-05-18"},{"type":"extras","title":"Added - Changelog","doc":"- A new `check_rate_inc` function, which allows the caller to specify the\n  integer with which to increment the bucket by. This is useful for limiting\n  APIs which have some notion of \"cost\" per call.","ref":"changelog.html#added"},{"type":"extras","title":"4.0.0 - 2018-04-23 - Changelog","doc":"","ref":"changelog.html#4-0-0-2018-04-23"},{"type":"extras","title":"Changed - Changelog","doc":"- Use a worker-pool for the backend (via poolboy),\n  this avoids bottle-necking all traffic through a single hammer\n  process, thus improving throughput for the system overall","ref":"changelog.html#changed"},{"type":"extras","title":"Added - Changelog","doc":"- New configuration options for backends:\n  - `:pool_size`, determines the number of workers in the pool (default 4)\n  - `:pool_max_overflow`, maximum extra workers to be spawned when the\n    system is under pressure (default 4)\n- Multiple instances of the same backend! You can now have two ETS backends,\n  fifteen Redis's, whatever you want","ref":"changelog.html#added"},{"type":"extras","title":"3.0.0 - 2018-02-20 - Changelog","doc":"","ref":"changelog.html#3-0-0-2018-02-20"},{"type":"extras","title":"Changed - Changelog","doc":"- Require elixir >= 1.6\n- Use a more sane supervision tree structure","ref":"changelog.html#changed"},{"type":"extras","title":"2.1.0 2017-11-25 - Changelog","doc":"","ref":"changelog.html#2-1-0-2017-11-25"},{"type":"extras","title":"Changed - Changelog","doc":"- Add option to use more than one backend\n- Add option to suppress all logging","ref":"changelog.html#changed"},{"type":"extras","title":"2.0.0 - 2017-09-24 - Changelog","doc":"","ref":"changelog.html#2-0-0-2017-09-24"},{"type":"extras","title":"Changed - Changelog","doc":"- New, simpler API\n  - No longer need to start backend processes manually\n  - Call `Hammer.check_rate` directly, rather than `use`ing a macro\n- Hammer is now an OTP application, configured via `Mix.Config`","ref":"changelog.html#changed"},{"type":"extras","title":"1.0.0 - 2017-08-22 - Changelog","doc":"","ref":"changelog.html#1-0-0-2017-08-22"},{"type":"extras","title":"Added - Changelog","doc":"- Formalise backend API in `Hammer.Backend` behaviour","ref":"changelog.html#added"},{"type":"extras","title":"0.2.1 - 2017-08-10 - Changelog","doc":"","ref":"changelog.html#0-2-1-2017-08-10"},{"type":"extras","title":"Changed - Changelog","doc":"- Minor fixes","ref":"changelog.html#changed"},{"type":"extras","title":"Hammer","doc":"# Hammer\n\n[![Build Status](https://github.com/ExHammer/hammer/actions/workflows/ci.yml/badge.svg)](https://github.com/ExHammer/hammer/actions/workflows/ci.yml)\n[![Hex.pm](https://img.shields.io/hexpm/v/hammer.svg)](https://hex.pm/packages/hammer)\n[![Documentation](https://img.shields.io/badge/documentation-gray)](https://hexdocs.pm/hammer)\n[![Total Download](https://img.shields.io/hexpm/dt/hammer.svg)](https://hex.pm/packages/hammer)\n[![License](https://img.shields.io/hexpm/l/hammer.svg)](https://github.com/ExHammer/hammer/blob/master/LICENSE.md)\n\n**Hammer** is a rate-limiter for Elixir with pluggable storage backends. Hammer enables users to set limits on actions performed within specified time intervals, applying per-user or global limits on API requests, file uploads, and more.\n\n---\n\n> [!NOTE]\n>\n> This README is for the unreleased master branch, please reference the [official documentation on hexdocs](https://hexdocs.pm/hammer) for the latest stable release.\n\n---","ref":"readme.html"},{"type":"extras","title":"Installation - Hammer","doc":"Hammer is [available in Hex](https://hex.pm/packages/hammer). Install by adding `:hammer` to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:hammer, \"~> 7.0\"}\n  ]\nend\n```","ref":"readme.html#installation"},{"type":"extras","title":"Available Backends - Hammer","doc":"Atomic backends are single-node rate limiting but will be the fastest option.\n\n- [Hammer.ETS](https://hexdocs.pm/hammer/Hammer.ETS.html) (default, can be [distributed](./guides/distributed-ets.md))\n- [Hammer.Atomic](https://hexdocs.pm/hammer/Hammer.Atomic.html)\n- [Hammer.Redis](https://github.com/ExHammer/hammer-backend-redis)\n- [Hammer.Mnesia](https://github.com/ExHammer/hammer-backend-mnesia)","ref":"readme.html#available-backends"},{"type":"extras","title":"Available Algorithms: - Hammer","doc":"Each backend supports multiple algorithms. Not all of them are available for all backends. The following table shows which algorithms are available for which backends.\n\n| Algorithm | Backend |\n| --------- | ------- |\n| [Hammer.Atomic.FixWindow](https://hexdocs.pm/hammer/Hammer.Atomic.FixWindow.html) | [Hammer.Atomic](https://hexdocs.pm/hammer/Hammer.Atomic.html) |\n| [Hammer.Atomic.LeakyBucket](https://hexdocs.pm/hammer/Hammer.Atomic.LeakyBucket.html) | [Hammer.Atomic](https://hexdocs.pm/hammer/Hammer.Atomic.html) |\n| [Hammer.Atomic.TokenBucket](https://hexdocs.pm/hammer/Hammer.Atomic.TokenBucket.html) | [Hammer.Atomic](https://hexdocs.pm/hammer/Hammer.Atomic.html) |\n| [Hammer.ETS.FixWindow](https://hexdocs.pm/hammer/Hammer.ETS.FixWindow.html) | [Hammer.ETS](https://hexdocs.pm/hammer/Hammer.ETS.html) |\n| [Hammer.ETS.LeakyBucket](https://hexdocs.pm/hammer/Hammer.ETS.LeakyBucket.html) | [Hammer.ETS](https://hexdocs.pm/hammer/Hammer.ETS.html) |\n| [Hammer.ETS.TokenBucket](https://hexdocs.pm/hammer/Hammer.ETS.TokenBucket.html) | [Hammer.ETS](https://hexdocs.pm/hammer/Hammer.ETS.html) |\n| [Hammer.ETS.SlidingWindow](https://hexdocs.pm/hammer/Hammer.ETS.SlidingWindow.html) | [Hammer.Redis](https://hexdocs.pm/hammer/Hammer.ETS.html) |\n| [Hammer.Redis.FixedWindow](https://hexdocs.pm/hammer/Hammer.Redis.FixedWindow.html) | [Hammer.Redis](https://hexdocs.pm/hammer/Hammer.Redis.html) |","ref":"readme.html#available-algorithms"},{"type":"extras","title":"Default Algorithm - Hammer","doc":"By default, Hammer backends use the **fixed window counter** to track actions within set time windows, resetting the count at the start of each new window. For example, with a limit of 10 uploads per minute, a user could upload up to 10 files between 12:00:00 and 12:00:59, and up to 10 more between 12:01:00 and 12:01:59. Notice that the user can upload 20 videos in a second if the uploads are timed at the window edges. If this is an issue, it can be worked around with a \"bursty\" counter which can be implemented with the current API by making two checks, one for the original interval with the total limit, and one for a shorter interval with a fraction of the limit. That would smooth out the number of requests allowed.","ref":"readme.html#default-algorithm"},{"type":"extras","title":"Algorithm Comparison - Hammer","doc":"Here's a comparison of the different rate limiting algorithms to help you choose:\n\n### [Fixed Window](https://hexdocs.pm/hammer/Hammer.Atomic.FixWindow.html)\n- Simplest implementation with lowest overhead\n- Good for basic rate limiting with clear time boundaries\n- Potential edge case: Up to 2x requests possible at window boundaries\n- Best for: Basic API limits where occasional bursts are acceptable\n\n### [Leaky Bucket](https://hexdocs.pm/hammer/Hammer.Atomic.LeakyBucket.html)\n- Provides smooth, consistent request rate\n- Requests \"leak\" out at constant rate\n- Good for traffic shaping and steady throughput\n- Best for: Network traffic control, queue processing\n\n### [Token Bucket](https://hexdocs.pm/hammer/Hammer.Atomic.TokenBucket.html)\n- Allows controlled bursts while maintaining average rate\n- Tokens regenerate at fixed rate\n- More flexible than fixed windows\n- Best for: APIs needing burst tolerance, gaming mechanics\n\n### [Sliding Window](https://hexdocs.pm/hammer/Hammer.ETS.SlidingWindow.html)\n- Most precise rate limiting\n- No boundary conditions like fixed windows\n- Higher overhead than other algorithms\n- Best for: Strict rate enforcement, critical systems\n\nSelection Guide:\n- Need simple implementation? → Fixed Window\n- Need smooth output rate? → Leaky Bucket\n- Need burst tolerance? → Token Bucket\n- Need precise limits? → Sliding Window","ref":"readme.html#algorithm-comparison"},{"type":"extras","title":"Creating a Rate Limiter - Hammer","doc":"- **Limit:** Maximum number of actions allowed in a window.\n- **Scale:** Duration of the time window (in milliseconds).\n- **Key:** Unique identifier (e.g., user ID) to scope the rate limiting.","ref":"readme.html#creating-a-rate-limiter"},{"type":"extras","title":"Example Usage - Hammer","doc":"```elixir\ndefmodule MyApp.RateLimit do\n  use Hammer, backend: :ets\nend\n\nMyApp.RateLimit.start_link()\n\nuser_id = 42\nkey = \"upload_video:#{user_id}\"\nscale = :timer.minutes(1)\nlimit = 3\n\ncase MyApp.RateLimit.hit(key, scale, limit) do\n  {:allow, _count} ->\n    # upload the video\n    :ok\n\n  {:deny, retry_after} ->\n    # deny the request\n    {:error, :rate_limit, _message = \"try again in #{retry_after}ms\"}\nend\n```","ref":"readme.html#example-usage"},{"type":"extras","title":"Benchmarks - Hammer","doc":"See the [BENCHMARKS.md](https://github.com/ExHammer/hammer/blob/master/BENCHMARKS.md) for more details.","ref":"readme.html#benchmarks"},{"type":"extras","title":"Acknowledgements - Hammer","doc":"Hammer was originally inspired by the [ExRated](https://github.com/grempe/ex_rated) library, by [grempe](https://github.com/grempe).","ref":"readme.html#acknowledgements"},{"type":"extras","title":"License - Hammer","doc":"Copyright (c) 2023 June Kelly\nCopyright (c) 2023-2024 See [CONTRIBUTORS.md](https://github.com/ExHammer/hammer/blob/master/CONTRIBUTORS.md)\n\nThis library is MIT licensed. See the [LICENSE](https://github.com/ExHammer/hammer/blob/master/LICENSE.md) for details.","ref":"readme.html#license"},{"type":"extras","title":"Tutorial","doc":"# Tutorial\n\nHammer is a rate limiting library for Elixir that can help you control the frequency of specific actions in your application, such as limiting API requests, login attempts, or file uploads. This tutorial will guide you through setting up Hammer, defining a rate limiter, and applying rate limiting in your app.","ref":"tutorial.html"},{"type":"extras","title":"Installation - Tutorial","doc":"Add Hammer as a dependency in `mix.exs`:\n\n```elixir\ndef deps do\n  [{:hammer, \"~> 7.0.0\"}]\nend\n```\n\nThen, run:\n\n```console\n$ mix deps.get\n```","ref":"tutorial.html#installation"},{"type":"extras","title":"Core Concepts - Tutorial","doc":"When rate-limiting an action, you specify a maximum number of allowed occurrences (the limit) within a certain time frame (the scale). For example, you might allow only 5 login attempts per minute for each user. The limit is typically enforced based on a unique identifier (like a user ID or IP address) but can also be applied globally.\n\nIn Hammer:\n- `limit` is the maximum number of actions permitted.\n- `scale` is the time period (in milliseconds) for that limit.\n- `key` is a unique identifier for the rate limit, combining the action name with a user identifier (like \"login_attempt:42\" for user 42) is a common approach.\n\nHammer uses a fixed window counter approach. It divides time into fixed-size windows of `scale` size and counts the number of requests in each window, blocking any requests that exceed the `limit`.","ref":"tutorial.html#core-concepts"},{"type":"extras","title":"Usage - Tutorial","doc":"To use Hammer, you need to:\n\n- Define a rate limiter module.\n- Add the Hammer backend to your application's supervision tree.\n\nIn this example, we'll use the Hammer.ETS backend, which stores data in an in-memory ETS table.","ref":"tutorial.html#usage"},{"type":"extras","title":"Step 1: Define a Rate Limiter - Tutorial","doc":"First, define a rate limiter module in your application. Use the `Hammer` module with your chosen backend and configure options as needed:\n\n```elixir\ndefmodule MyApp.RateLimit do\n  use Hammer, backend: :ets\nend\n```\n\nHere:\n- `:backend` specifies the storage backend (`:ets` for in-memory storage, `Hammer.Redis` for Redis, etc.).","ref":"tutorial.html#step-1-define-a-rate-limiter"},{"type":"extras","title":"Step 2: Start the Rate Limiter - Tutorial","doc":"Add the rate limiter to your application's supervision tree or start it manually by calling `MyApp.RateLimit.start_link/1` with any runtime options:\n\n```elixir\nMyApp.RateLimit.start_link(clean_period: :timer.minutes(1))\n```\n\n- `:clean_period` is an optional parameter for `:ets` backend that specifies how often to clean expired buckets in the ETS table.","ref":"tutorial.html#step-2-start-the-rate-limiter"},{"type":"extras","title":"Using the Rate Limiter - Tutorial","doc":"With the rate limiter running, you can use `hit/3` or `hit/4` to enforce rate limits.","ref":"tutorial.html#using-the-rate-limiter"},{"type":"extras","title":"Example: Basic Rate Limit Check - Tutorial","doc":"Suppose you want to limit file uploads to 10 per minute per user.\n\n```elixir\nuser_id = 42\nkey = \"upload_file:#{user_id}\"\nscale = :timer.minutes(1)\nlimit = 10\n\ncase MyApp.RateLimit.hit(key, scale, limit) do\n  {:allow, _current_count} -> # proceed with file upload\n  {:deny, _ms_until_next_window} -> # deny the request\nend\n```","ref":"tutorial.html#example-basic-rate-limit-check"},{"type":"extras","title":"Customizing Rate Increments - Tutorial","doc":"If you want to specify a custom increment—useful when each action has a \"cost\"—you can use `hit/4`. Here's an example for a bulk upload scenario:\n\n```elixir\nuser_id = 42\nkey = \"upload_file:#{user_id}\"\nscale = :timer.minutes(1)\nlimit = 10\nnumber_of_files = 3\n\ncase MyApp.RateLimit.hit(key, scale, limit, number_of_files) do\n  {:allow, _current_count} -> # upload all files\n  {:deny, _ms_until_next_window} -> # deny the request\nend\n```","ref":"tutorial.html#customizing-rate-increments"},{"type":"extras","title":"Using Hammer as a Plug in Phoenix - Tutorial","doc":"you can easily use Hammer as a plug by using the controller plug in Phoenix:\n\n```elixir\nplug :rate_limit_videos when action in ...\n\ndefp rate_limit_videos(conn, _opts) do\n  user_id = conn.assigns.current_user.id\n  key = \"videos:#{user_id}\"\n  scale = :timer.minutes(1)\n  limit = 10\n\n  case MyApp.RateLimit.hit(key, scale, limit) do\n    {:allow, _count} ->\n      conn\n\n    {:deny, retry_after} ->\n      conn\n      |> put_resp_header(\"retry-after\", Integer.to_string(div(retry_after, 1000)))\n      |> send_resp(429, [])\n      |> halt()\n  end\nend\n```\n\nOr you could add it to your endpoint:\n\n```elixir\ndefmodule MyAppWeb.Endpoint do\n  use Phoenix.Endpoint\n\n  plug RemoteIP\n  plug :rate_limit\n\n  # ...\n\n  defp rate_limit(conn, _opts) do\n    key = \"web_requests:#{:inet.ntoa(conn.remote_ip)}\"\n    scale = :timer.minutes(1)\n    limit = 1000\n\n    case MyApp.RateLimit.hit(key, scale, limit) do\n      {:allow, _count} ->\n        conn\n\n      {:deny, retry_after} ->\n        retry_after_seconds = div(retry_after, 1000)\n\n        conn\n        |> put_resp_header(\"retry-after\", Integer.to_string(retry_after_seconds))\n        |> send_resp(429, [])\n        |> halt()\n    end\n  end\nend\n```","ref":"tutorial.html#using-hammer-as-a-plug-in-phoenix"},{"type":"extras","title":"Using Hammer with Redis - Tutorial","doc":"To persist rate-limiting data across multiple nodes, you can use the Redis backend. Install the `Hammer.Redis` backend and update your rate limiter configuration:\n\n```elixir\ndefmodule MyApp.RateLimit do\n  use Hammer, backend: Hammer.Redis\nend\n```\n\nThen, start the rate limiter pool with Redis configuration:\n\n```elixir\nMyApp.RateLimit.start_link(host: \"redix.myapp.com\")\n```\n\nConfiguration options are the same as [Redix](https://hexdocs.pm/redix/Redix.html#start_link/1), except for `:name`, which comes from the module definition.","ref":"tutorial.html#using-hammer-with-redis"},{"type":"extras","title":"Distributed Rate Limiter with ETS Backend","doc":"# Distributed Rate Limiter with ETS Backend\n\nThis example implements a distributed, eventually consistent rate limiter using Phoenix.PubSub for broadcasting each hit across nodes and a local ETS backend to manage rate-limiting counters. This setup is useful when you need to limit the number of actions (e.g., requests) across multiple nodes in a cluster.\n\nBased on [HexpmWeb.RateLimitPubSub.](https://github.com/hexpm/hexpm/blob/main/lib/hexpm_web/rate_limit_pub_sub.ex)\n\n```elixir\ndefmodule MyApp.RateLimit do\n  @moduledoc \"\"\"\n  Distributed, eventually consistent rate limiter using `Phoenix.PubSub` and `Hammer`.\n  \n  This module provides a rate-limiting mechanism for requests using a distributed,\n  eventually consistent approach. It combines local in-memory counting with a \n  broadcasting mechanism to keep counters in sync across nodes in a cluster.\n  \"\"\"\n\n  # Checks rate locally and broadcasts the hit to other nodes to synchronize.\n  def hit(key, scale, limit, increment \\\\ 1) do\n    :ok = broadcast({:inc, key, scale, increment})\n    Local.hit(key, scale, limit, increment)\n  end\n\n  defmodule Local do\n    @moduledoc false\n    use Hammer, backend: :ets\n    # This inner module handles local hit counting via Hammer with ETS as a backend.\n  end\n\n  defmodule Listener do\n    @moduledoc false\n    use GenServer\n\n    # Starts the listener process, subscribing to the specified PubSub topic.\n    # This process will listen for `:inc` messages to keep local counters in sync.\n\n    @doc false\n    def start_link(opts) do\n      pubsub = Keyword.fetch!(opts, :pubsub)\n      topic = Keyword.fetch!(opts, :topic)\n      GenServer.start_link(__MODULE__, {pubsub, topic})\n    end\n\n    @impl true\n    def init({pubsub, topic}) do\n      :ok = Phoenix.PubSub.subscribe(pubsub, topic)\n      {:ok, []}\n    end\n\n    # Handles remote `:inc` messages by updating the local counter.\n\n    @impl true\n    def handle_info({:inc, key, scale, increment}, state) do      \n      _count = Local.inc(key, scale, increment)\n      {:noreply, state}\n    end\n  end\n\n  @pubsub MyApp.PubSub\n  @topic \"__ratelimit\"\n\n  # Sends a message to other nodes in the cluster to synchronize rate-limiting information.\n  defp broadcast(message) do\n    {:ok, {Phoenix.PubSub.PG2, adapter_name}} = Registry.meta(@pubsub, :pubsub)\n    adapter_name.broadcast(adapter_name, @topic, message)\n  end\n\n  def child_spec(opts) do\n    %{\n      id: __MODULE__,\n      start: {__MODULE__, :start_link, [opts]},\n      type: :supervisor\n    }\n  end\n\n  # Wraps the local Hammer counter and the listener processes under a single supervisor.\n  def start_link(opts) do\n    children = [{Local, opts}, {Listener, pubsub: @pubsub, topic: @topic}]\n    Supervisor.start_link(children, strategy: :one_for_one)\n  end\nend\n```\n\nTo be covered:\n- new nodes joining have empty counters\n- net split recovery\n- dns_cluster","ref":"distributed-ets.html"},{"type":"extras","title":"Upgrading to Hammer V7","doc":"# Upgrading to Hammer V7","ref":"upgrade-v7.html"},{"type":"extras","title":"Elixir and Erlang/OTP Compatibility - Upgrading to Hammer V7","doc":"* Hammer v7 requires Elixir 1.14 and Erlang/OTP 25 at a minimum.\n* We recommend using the latest Elixir and Erlang/OTP versions.","ref":"upgrade-v7.html#elixir-and-erlang-otp-compatibility"},{"type":"extras","title":"Changes to your Project - Upgrading to Hammer V7","doc":"* Update your `mix.exs` to depend on version `7.0.0` of Hammer.\n\n```elixir\ndef deps do\n  [\n    ...\n    {:hammer, \"~> 7.0.0\"}\n    ...\n    ]\nend\n```","ref":"upgrade-v7.html#changes-to-your-project"},{"type":"extras","title":"Define a Rate Limiter - Upgrading to Hammer V7","doc":"First, define a rate limiter module in your application. Use the `Hammer` module with your chosen backend and configure options as needed:\n\n```elixir\ndefmodule MyApp.RateLimit do\n  use Hammer, backend: :ets\nend\n```\n\nThis would setup the rate limiter using the `Hammer.ETS` backend. See the [Tutorial](./Tutorial.md) guide for more information on other backends.","ref":"upgrade-v7.html#define-a-rate-limiter"},{"type":"extras","title":"Update your Application Supervisor - Upgrading to Hammer V7","doc":"* Pick up the value in your config file for `cleanup_interval_ms`.\n* remove the `config` lines for `Hammer` as they are no longer needed in all of the `config/*.exs` files.\n* In your `application.ex` file, add the following line to start the rate limiter:\n\n```elixir\ndef start(_type, _args) do\n\n  children = [\n    ...\n    {MyApp.RateLimit, [clean_period: 60_000]}\n    ...\n  ]\n\n  Supervisor.start_link(children, strategy: :one_for_one)\nend\n```","ref":"upgrade-v7.html#update-your-application-supervisor"},{"type":"extras","title":"Change to Backend Configuration - Upgrading to Hammer V7","doc":"We have simplified the backend API. `Hammer.inc/4` combines the functionality of `Hammer.check_rate` and `Hammer.check_rate_inc` now.\n\n* Remapped all the  `Hammer.check_rate/3` and  `Hammer.check_rate/4` to `Hammer.inc/4`.\n* Remapped all the  `Hammer.check_rate_inc/4` and  `Hammer.check_rate_inc/5` to `Hammer.inc/4`.\n* for the `Hammer.delete_buckets`, you need to remove them as there no true replacement. You could potentially use `Hammer.ETS.set/1` to reset specific key\n* for the `Hammer.make_rate_checker`, you need to remove them as there no replacement.","ref":"upgrade-v7.html#change-to-backend-configuration"},{"type":"extras","title":"Changes to the Hammer.Plug - Upgrading to Hammer V7","doc":"* The `Hammer.Plug` has been removed. Remove any references to it in your code.\n* Migrate it by using regular Phoenix plugs in either a controller plug or an endpoint plug. See the [Tutorial](./Tutorial.md) guide for some examples.","ref":"upgrade-v7.html#changes-to-the-hammer-plug"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,52,46,50]}}